unit DemoComp;

{ Demo how to draw line chart with preview and printing capabilities. }
{ You can use this as normal component or with PaintTo method you can
  setup print previewing and printing to printer. }

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs;

type
  TDemo = class(TGraphicControl)
  private
    { Canvas where to draw. Depending on situation it's assign to
      off-screen bitmap, some component's canvas or printers canvas. }
    fCanvas: TCanvas;
    { Offscreen bitmap for drawing. Finally contents of the off-screen
      bitmap is copied to component's own canvas. }
    fOffScreen: TBitmap;
    { These variables defines size and position of the actual drawing area. }
    fLeft, fTop, fRight, fBottom, fHeight, fWidth: longint;
    { These variables defines size and position of chart area inside
      the actual drawing area. }
    fX0, fY0, fX1, fY1: longint;
    { Titles for chart, x-axis and y-axis. }
    fTitle, fXTitle, fYTitle: string;
    { Line widths. Values are depending on the size of canvas where to draw. }
    fThinLine, fNormalLine: longint;
    { Data array for points.}
    fData: array[0..10] of integer;

    procedure DrawPicture;
    procedure DrawTitles;
    procedure DrawXandYAxis;
    procedure DrawLine;
    procedure SetTitle(value: string);
    procedure SetXTitle(value: string);
    procedure SetYTitle(value: string);
  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure Paint; override;
    procedure PaintNoErase;
    procedure PaintTo(c: TCanvas; x, y, w, h: integer);
    function  ValuesToPoint(x, y: longint; var p: TPoint): boolean;
  published
    property Align;
    property Title:  string read fTitle write SetTitle;
    property XTitle: string read fXTitle write SetXTitle;
    property YTitle: string read fYTitle write SetYTitle;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Samples', [TDemo]);
end;

constructor TDemo.Create(AOwner: TComponent);
var i: integer;
begin
  inherited Create(AOwner);
  Parent := TWinControl(AOwner);
  fOffScreen := TBitmap.Create;
  fOffScreen.Height := Screen.Height;
  fOffScreen.Width := Screen.Width;
  fTitle := 'Title';
  fXTitle := 'X-axis';
  fYTitle := 'Y-axis';
  Width := 210;
  Height := 150;
  Randomize;
  for i := 0 to 10 do
    fData[i] := 1 + Random(9);
end;

destructor TDemo.Destroy;
begin
  fOffScreen.Free;
  inherited Destroy;
end;

{ Paint: Default painting procedure. Uses off-screen bitmap for
  flicker-free drawing. }

procedure TDemo.Paint;
begin
  fLeft := 0;
  fTop := 0;
  fRight := Width;
  fBottom := Height;
  fWidth := Width;
  fHeight := Height;
  fOffScreen.Canvas.Pen.Color := clWhite;
  fOffScreen.Canvas.Brush.Color := clWhite;
  fOffScreen.Canvas.Rectangle(0, 0, Width, Height);
  fCanvas := fOffScreen.Canvas;
  DrawPicture;
  Canvas.CopyMode := cmSrcCopy;
  Canvas.CopyRect(Rect(0,0,Width,Height), fOffScreen.Canvas,
                  Rect(0,0,Width,Height));
end;

{ PaintNoErase: Calling this routine from parent form's OnPaint
  event causes this component's area in form to be validated so
  Windows' default painting procedure do not erase this area
  before drawing. Result is flicker-free component. This requires
  that entire area off component must be draw inside component.
  That's why there's Rectangle call in Paint. }

procedure TDemo.PaintNoErase;
var pRec: PRect;
begin
  Paint;
  GetMem(pRec, SizeOf(TRect));
  pRec^.Left   := Left;
  pRec^.Top    := Top;
  pRec^.Right  := Left + Width;
  pRec^.Bottom := Top + Height;
  ValidateRect(Parent.Handle, pRec);
  FreeMem(pRec, SizeOf(TRect));
end;

procedure TDemo.PaintTo(c: TCanvas; x, y, w, h: integer);
{ PaintTo procedure draws area to given canvas using position
  and size passed as parameters. }
begin
  fCanvas := c;
  fLeft := x;
  fTop := y;
  fRight := x + w;
  fBottom := y + h;
  fWidth := w;
  fHeight := h;
  c.Pen.Color := clWhite;
  c.Brush.Color := clWhite;
  c.Brush.Style := bsSolid;
  c.Rectangle(fLeft, fTop, fRight, fBottom);
  DrawPicture;
end;

{ Here's actual painting procedure. It draw all objects like fonts,
  lines etc. relative to size of drawing area. }

procedure TDemo.DrawPicture;
begin
  fThinLine := fHeight div 200;
  if fThinLine = 0 then
    fThinLine := 1;
  fNormalLine := fHeight div 100;
  if fNormalLine = 0 then
    fNormalLine := 1;
  DrawTitles;
  DrawXandYAxis;
  DrawLine;
end;

procedure TDemo.DrawTitles;
var
  hOrigFont, hXFont, hYFont, hTFont: hFont;
  XTextHeight, YTextHeight, TitleHeight: longint;
  x, y: longint;
begin
  fCanvas.Font.Color := clBlue;

  { Creating own fonts. Note that hYFont is rotated 90 degrees. }
  hXFont := CreateFont(fHeight div 25, fWidth div 70, 0, 0, FW_REGULAR, 0, 0, 0, 1,
                       OUT_DEFAULT_PRECIS, 0, PROOF_QUALITY, FF_DONTCARE,
                      'Arial');
  hYFont := CreateFont(fHeight div 25, fWidth div 70, 900, 0, FW_REGULAR, 0, 0, 0, 1,
                       OUT_DEFAULT_PRECIS, 0, PROOF_QUALITY, FF_DONTCARE,
                      'Arial');
  hTFont := CreateFont(fHeight div 15, fWidth div 40, 0, 0, FW_REGULAR, 0, 0, 0, 1,
                       OUT_DEFAULT_PRECIS, 0, PROOF_QUALITY, FF_DONTCARE,
                      'Arial');

  { When first time changing font must original font save. }
  hOrigFont := SelectObject(fCanvas.Handle, hXFont);
  XTextHeight := fCanvas.TextHeight('M');
  SelectObject(fCanvas.Handle, hYFont);
  YTextHeight := fCanvas.TextHeight('M');
  SelectObject(fCanvas.Handle, hTFont);
  TitleHeight := fCanvas.TextHeight('M');

  { Calculate the drawing area for chart (excluding the title area) }
  fY0 := fTop + TitleHeight + 4 * fNormalLine;
  fY1 := fTop + fHeight - XTextHeight - 6 * fNormalLine;
  fX0 := fLeft + YTextHeight + 6 * fNormalLine;
  fX1 := fLeft + fWidth - 6 * fNormalLine;

  { Print title texts }
  SelectObject(fCanvas.Handle, hYFont);
  x := fLeft + 2 * fNormalLine;
  y := fY0 + (fY1 - fY0 + fCanvas.TextWidth(fYTitle)) div 2;
  fCanvas.TextOut(x, y, fYTitle);

  SelectObject(fCanvas.Handle, hXFont);
  x := fX0 + (fX1 - fX0 - fCanvas.TextWidth(fXTitle)) div 2;
  y := fY1 + 4 * fNormalLine;
  fCanvas.TextOut(x, y, fXTitle);

  SelectObject(fCanvas.Handle, hTFont);
  x := fLeft + (fWidth - fCanvas.TextWidth(fTitle)) div 2;
  y := fTop + 2 * fNormalLine;
  fCanvas.TextOut(x, y, fTitle);

  { Restore original font and delete own ones. }
  SelectObject(fCanvas.Handle, hOrigFont);
  DeleteObject(hXFont);
  DeleteObject(hYFont);
  DeleteObject(hTFont);
end;

procedure TDemo.DrawXandYAxis;
var iLoop: integer;
begin
  with fCanvas do begin
    Pen.Width := fNormalLine;
    Pen.Color := clBlack;
    { Draw Y-axis and X-axis.}
    MoveTo(fX0, fY0);
    LineTo(fX0, fY1);
    LineTo(fX1, fY1);
    { Draw the arrow heads. }
    MoveTo(fX0 - 2 * fNormalLine, fY0 + 4 * fNormalLine);
    LineTo(fX0, fY0);
    LineTo(fX0 + 2 * fNormalLine, fY0 + 4 * fNormalLine);
    MoveTo(fX1 - 4 * fNormalLine, fY1 - 2 * fNormalLine);
    LineTo(fX1, fY1);
    LineTo(fX1 - 4 * fNormalLine, fY1 + 2 * fNormalLine);
    { Draw gridlines. }
    Pen.Style := psDot;
    Pen.Width := fThinLine;
    Pen.Color := clSilver;
    for iLoop := 1 to 9 do begin
      MoveTo(fX0, fY1 - iLoop * (fY1 - fY0) div 10);
      LineTo(fX1, fY1 - iLoop * (fY1 - fY0) div 10);
      MoveTo(fX0 + iLoop * (fX1 - fX0) div 10, fY0);
      LineTo(fX0 + iLoop * (fX1 - fX0) div 10, fY1);
    end;
    { Draw markers on x- and y-axis. }
    Pen.Style := psSolid;
    Pen.Width := fNormalLine;
    Pen.Color := clBlack;
    for iLoop := 1 to 9 do begin
      MoveTo(fX0 - fNormalLine, fY1 - iLoop * (fY1 - fY0) div 10);
      LineTo(fX0 + fNormalLine, fY1 - iLoop * (fY1 - fY0) div 10);
      MoveTo(fX0 + iLoop * (fX1 - fX0) div 10, fY1 - fNormalLine);
      LineTo(fX0 + iLoop * (fX1 - fX0) div 10, fY1 + fNormalLine);
    end;
  end;
end;

procedure TDemo.DrawLine;
var p: TPoint;
    x, y: integer;
begin
  fCanvas.Pen.Width := fThinLine;
  fCanvas.Pen.Color := clRed;
  fCanvas.Pen.Style := psSolid;
  ValuesToPoint(0, fData[0], p);
  fCanvas.MoveTo(p.x, p.y);
  for x := 1 to 10 do begin
    ValuesToPoint(x, fData[x], p);
    fCanvas.LineTo(p.x, p.y);
  end;
end;

function TDemo.ValuesToPoint(x, y: longint; var p: TPoint): boolean;
begin
  p.x := fX0 + (x * (fX1 - fX0) div 10);
  p.y := fY1 + (y * (fY0 - fY1) div 10);
  ValuesToPoint := (p.x in [fX0..fX1]) and (p.y in [fY1..fY0]);
end;

procedure TDemo.SetTitle(value: string);
begin
  if value <> fTitle then begin
    fTitle := value;
    Invalidate;
  end;
end;

procedure TDemo.SetXTitle(value: string);
begin
  if value <> fXTitle then begin
    fXTitle := value;
    Invalidate;
  end;
end;

procedure TDemo.SetYTitle(value: string);
begin
  if value <> fYTitle then begin
    fYTitle := value;
    Invalidate;
  end;
end;

end.

