unit Logfiles;

{  **************************************************
   *   TLogFile component                           *
   *   February 2, 1996                             *
   *   Concocted by Keith Rome                      *
   *                                                *
   *     TLogFile provides an interface to a        *
   *   simple application 'event log'.  By just     *
   *   specifying a file name, it handles the       *
   *   tedious particulars of implementing a log    *
   *   file.                                        *
   *                                                *
   *   Notice:  This Delphi component is basically  *
   *            FREEWARE.  Do whatever you want     *
   *            with it, but please give me credit  *
   *            if you use it.  This component is   *
   *            NOT for use in any commercial       *
   *            application without written permiss-*
   *            ion from myself.  If you find this  *
   *            component useful, or if you find a  *
   *            way to improve it, drop me a line.  *
   *            If you don't like it, tough, you'll *
   *            get over it.                        *
   *                                                *
   *   For information on the properties/methods/   *
   *   Events of this component, see the included   *
   *   Write document.                              *
   **************************************************  }

interface

uses
  SysUtils, WinTypes, WinProcs, Classes, Forms, StdCtrls;

type
  {Options for what to record in each event}
  TLogFileDisplayOptions = (ldoTime, ldoSeperator);
  TLogFileRecordOptions = set of TLogFileDisplayOptions;

  {Options for what application events to automatically monitor and record}
  TLogFileLogOptions = (loCreate, loDestroy, loMoveFile, loCreateFile, loClearFile, loViewFile);
  TLogFileLogSettings = set of TLogFileLogOptions;

  {Event procedure type for OnLogEvent}
  TLogFileEvent = procedure(Sender : TObject; EventString : string) of object;

  TLogFile = class(TComponent)
  private
    {private data fields}
    FLogFileName: string;    {Full path to event log file}
    FEditor: string;       {Executable name of editor to use for viewing log file}
    FRecordOptions: TLogFileRecordOptions;  {set of options used to determine what to record in file}
    FLogOptions: TLogFileLogSettings;  {set of options used to determine what standard events to record}
    FSeperatorChar: char;  {Single character used to seperate date/time and message in event log}
    FEventOutput: TListBox;  {Optional pointer to a list box which can receive messages output to log file}

    {private event handler pointer field}
    FOnLogEvent: TLogFileEvent;  

    {private accessor functions for properties}
    procedure SetLogFileName(NewName: string);
    procedure SetEditor(Value: string);
    procedure SetRecordOptions(Value: TLogFileRecordOptions);
    procedure SetLogOptions(Value: TLogFileLogSettings);
    procedure SetSeperatorChar(Value: Char);
  protected
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;  
    destructor Destroy; override;
    procedure LogEvent(Event: string);  {User can log a non-standard message with this}
    procedure EditLog;                  {User can view entire log file in a seperate editor}
    procedure ClearLog;                 {User can clear contents of current log file}
  published
    property LogFileName : string
       read FLogFileName
       write SetLogFileName;
    property Editor :string
       read FEditor
       write SetEditor;
    property RecordOptions : TLogFileRecordOptions
       read FRecordOptions
       write SetRecordOptions
       default [ldoTime, ldoSeperator];
    property LogOptions : TLogFileLogSettings
       read FLogOptions
       write SetLogOptions
       default [loCreate, loDestroy, loMoveFile, loCreateFile, loClearFile, loViewFile];
    property SeperatorChar : Char
       read FSeperatorChar
       write SetSeperatorChar
       default '!';
    property EventOutput : TListBox
       read FEventOutput
       write FEventOutput;
    property OnLogEvent : TLogFileEvent
       read FOnLogEvent
       write FOnLogEvent;
  end;

procedure Register;

implementation

{################### BEGIN TLogFile #####################}
{First, set some default properties}
constructor TLogFile.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FEditor := 'NOTEPAD.EXE';
  FLogFileName := 'C:\Event.Log';
  FSeperatorChar := '!';
  FRecordOptions := [ldoTime, ldoSeperator];
  FLogOptions := [loCreate, loDestroy, loMoveFile, loCreateFile, loClearFile, loViewFile];
end;

{Here is where we can capture the shutdown of the TLogFile component}
destructor TLogFile.Destroy;
begin
  FEventOutput := nil;
  if (loDestroy in FlogOptions) and (not (csDesigning in ComponentState)) then
    LogEvent(Application.Title + ' terminated');
  inherited Destroy;
end;

{Here is where we can capture the creation of the TLogFile... i.e., application startup, for logging}
procedure TLogFile.Loaded;
begin
  inherited Loaded;
  {We don't want to log when in design mode}
  if (loCreate in FLogOptions) and (not (csDesigning in ComponentState)) then
    LogEvent(Application.Title + ' has been started');
end;

{The method actually used to log an event}
procedure TLogFile.LogEvent(Event: string);
var
  TmpEvent: string;
  FFileHandle: TextFile;
begin
  {Make sure the event log file already exists} 
  if not FileExists(FLogFileName) then
    begin
      FileClose(FileCreate(FLogFileName));
      if loCreateFile in FLogOptions then
        LogEvent('Created New Log File');
    end;
  
  {Build the string to log in file}
  TmpEvent := '';
  if ldoTime in FRecordOptions then
    TmpEvent := TmpEvent + DateTimeToStr(Now) + ' ';
  if ldoSeperator in FRecordOptions then
    TmpEvent := TmpEvent + FSeperatorChar + ' ';
  TmpEvent := TmpEvent + Event;

  {Actually log event to file}
  AssignFile(FFileHandle, FLogFileName);
  Append(FFileHandle);
  WriteLn(FFileHandle, TmpEvent);
  CloseFile(FFileHandle);

  {post event to TListBox, if set}
  if assigned(FEventOutput) then
    begin
      FEventOutput.Items.Add(TmpEvent);
      FEventOutput.ItemIndex := FEventOutput.Items.Count - 1;
    end;

  {call OnLogEvent event handler, if defined}
  if assigned(FOnLogEvent) then
    FOnLogEvent(Self, Event);
end;

{Method used to launch external editor}
procedure TLogFile.EditLog;
var
  TmpEditor : array[0..255] of char;
begin
  if loViewFile in FLogOptions then
    LogEvent('Log File loaded in external viewer - ' + FEditor);
  WinExec(StrPCopy(TmpEditor, FEditor + ' ' + FLogFileName), SW_NORMAL);
end;

{Method used to clear all events}
procedure TLogFile.ClearLog;
begin
  if FileExists(FLogFileName) then DeleteFile(FLogFileName);
  if assigned(FEventOutput) then
    FEventOutput.Clear;
  if loClearFile in FLogOptions then
    LogEvent('Log File Cleared');
end;

{Accessor methods for properties}
procedure TLogFile.SetLogFileName(NewName: string);
begin
  if NewName <> FLogFileName then
    FLogFileName := NewName;
end;

procedure TLogFile.SetEditor(Value: string);
begin
  if Value <> FEditor then
    FEditor := Value;
end;

procedure TLogFile.SetRecordOptions(Value: TLogFileRecordOptions);
begin
  if FRecordOptions <> Value then
    FRecordOptions := Value;
end;

procedure TLogFile.SetLogOptions(Value: TLogFileLogSettings);
begin
  if FLogOptions <> Value then
    FLogOptions := Value;
end;

procedure TLogFile.SetSeperatorChar(Value: Char);
begin
  if FSeperatorChar <> Value then
    FSeperatorChar := Value;
end;
{#################### END TLogFile ######################}


procedure Register;
begin
  RegisterComponents('The Good Package', [TLogFile]);
end;

end.
