unit Profile;

{---------------------------------------------------------------------------------}
{                                                                                 }
{ TProfileFile, TProfileSection, TProfileInteger, TProfileBoolean, TProfileString }
{                                                                                 }
{ Encapsulates .INI files.  At run time, the filename defaults to the executable  }
{ name with .INI substituted for .EXE.                                            }
{                                                                                 }
{ Freeware.  Use it at your own risk.                                             }
{                                                                                 }
{                                                                                 }
{ S Armstrong     13/1/96                                                         }
{ 100717.3531@compuserve.com                                                      }
{---------------------------------------------------------------------------------}

interface

uses
   Classes, IniFiles, Forms, SysUtils;

type
  TInvalidIdentException = class(Exception);

  TProfileFile = class(TComponent)
  private
    { Private declarations }
    FIniFile: TIniFile;
    FFilename: TFileName;
    procedure SetFilename(fn: TFileName);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create(Aowner: TComponent); override;
    destructor destroy;
    function  GetIniFile: TIniFile;
  published
    { Published declarations }
    property Filename: TFileName read FFilename write SetFilename;
  end;

  TProfileSection = class(TComponent)
  private
    { Private declarations }
    FProfileFile: TProfileFile;
    FSection: string;
  protected
    { Protected declarations }
     procedure SetName(const NewName: TComponentName); override;
  public
    { Public declarations }
  published
    { Published declarations }
    property ProfileFile: TProfileFile read FProfileFile write FProfileFile;
    property Section: string read FSection write FSection;
  end;

  TProfileIdent = class(TComponent)
  private
    { Private declarations }
    FProfileSection: TProfileSection;
    FIdent: string;
    procedure SetProfileSection(ps: TProfileSection);
    procedure SetIdent(id: string);
  protected
    { Protected declarations }
    procedure SetName(const NewName: TComponentName); override;
    procedure Update; virtual;
    function  ReadInteger(const ident: string; const default: integer): integer;
    procedure WriteInteger(const ident: string; const value: integer);
    function  ReadBoolean(const ident: string; const default: boolean): boolean;
    procedure WriteBoolean(const ident: string; const value: boolean);
    function  ReadString(const ident, default: String): String;
    procedure WriteString(const ident, value: String);
  public
    { Public declarations }
  published
    { Published declarations }
    property ProfileSection: TProfileSection read FProfileSection write SetProfileSection;
    property Ident: string read FIdent write SetIdent;
  end;

  TProfileString = class(TProfileIdent)
  private
    { Private declarations }
    FValue: String;
    FDefaultValue: String;
    procedure SetDefaultValue(v: String);
    procedure SetValue(v: String);
  protected
    { Protected declarations }
    procedure Update; override;
  public
    { Public declarations }
    constructor create(AOwner: TComponent); override;
    property Value: String read FValue write SetValue;
  published
    { Published declarations }
    property DefaultValue: String read FDefaultValue write SetDefaultValue;
  end;

  TProfileInteger = class(TProfileIdent)
  private
    { Private declarations }
    FValue: integer;
    FDefaultValue: integer;
    procedure SetDefaultValue(v: integer);
    procedure SetValue(v: integer);
  protected
    { Protected declarations }
    procedure Update; override;
  public
    { Public declarations }
    constructor create(AOwner: TComponent); override;
    property Value: integer read FValue write SetValue;
  published
    { Published declarations }
    property DefaultValue: integer read FDefaultValue write SetDefaultValue default 0;
  end;

  TProfileBoolean = class(TProfileIdent)
  private
    { Private declarations }
    FValue: boolean;
    FDefaultValue: boolean;
    procedure SetDefaultValue(v: boolean);
    procedure SetValue(v: boolean);
  protected
    { Protected declarations }
    procedure Update; override;
  public
    { Public declarations }
    constructor create(AOwner: TComponent); override;
    property Value: boolean read FValue write SetValue;
  published
    { Published declarations }
    property DefaultValue: boolean read FDefaultValue write SetDefaultValue default false;
  end;

  procedure Register;

implementation

{******************************************************************************}

procedure Register;
begin
  RegisterComponents('Profile', [TProfileFile, TProfileSection, TProfileString,
                                 TProfileInteger, TProfileBoolean]);
end;

{******************************************************************************}

constructor TProfileFile.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FFilename := '';
   if not (csDesigning in ComponentState) then
      SetFilename(ChangeFileExt(Application.ExeName,'.INI'));
end;

destructor TProfileFile.Destroy;
begin
   FIniFile.Free;
   inherited destroy;
end;

function TProfileFile.GetIniFile: TIniFile;
begin
   result := FIniFile;
end;

procedure TProfileFile.SetFilename(fn: TFileName);
begin
   if FFilename <> fn then begin
      FFilename := fn;
      FIniFile.Free;
      FIniFile := TIniFile.Create(FFilename);
   end;
end;

{******************************************************************************}

procedure TProfileSection.SetName(const NewName: TComponentName);
var
   same: boolean;
begin
   same := FSection = Name;
   inherited SetName(NewName);
   if same then begin
      FSection := Name;
   end;
end;

{******************************************************************************}

procedure TProfileIdent.SetProfileSection(ps: TProfileSection);
begin
   if FProfileSection <> ps then begin
      FProfileSection := ps;
      Update;
   end;
end;

procedure TProfileIdent.SetIdent(id: string);
begin
   if FIdent <> id then begin
      if Pos(id,'=') > 0 then
         raise TInvalidIdentException.Create(id + ' is an invalid ident name')
      else begin
         FIdent := id;
         Update;
      end;
   end;
end;

procedure TProfileIdent.SetName(const NewName: TComponentName);
var
   same: boolean;
begin
   same := FIdent = Name;
   inherited SetName(NewName);
   if same then
      FIdent := Name;
end;

procedure TProfileIdent.Update;
begin
   { allow overriding in descendents }
end;

function TProfileIdent.ReadInteger(const ident: string; const default: integer): integer;
begin
   if (ProfileSection <> nil) and
      (ProfileSection.ProfileFile <> nil) and
      (ProfileSection.ProfileFile.GetIniFile <> nil) then
      result := ProfileSection.ProfileFile.GetIniFile.ReadInteger(ProfileSection.section,ident,default)
   else
      result := default;
end;

procedure TProfileIdent.WriteInteger(const ident: string; const value: integer);
begin
   if (ProfileSection <> nil) and
      (ProfileSection.ProfileFile <> nil) and
      (ProfileSection.ProfileFile.GetIniFile <> nil) then
      ProfileSection.ProfileFile.GetIniFile.WriteInteger(ProfileSection.Section,ident,value);
end;

function TProfileIdent.ReadBoolean(const ident: string; const default: Boolean): Boolean;
begin
   if (ProfileSection <> nil) and
      (ProfileSection.ProfileFile <> nil) and
      (ProfileSection.ProfileFile.GetIniFile <> nil) then
      result := ProfileSection.ProfileFile.GetIniFile.ReadBool(ProfileSection.section,ident,default)
   else
      result := default;
end;

procedure TProfileIdent.WriteBoolean(const ident: string; const value: Boolean);
begin
   if (ProfileSection <> nil) and
      (ProfileSection.ProfileFile <> nil) and
      (ProfileSection.ProfileFile.GetIniFile <> nil) then
      ProfileSection.ProfileFile.GetIniFile.WriteBool(ProfileSection.Section,ident,value);
end;

function TProfileIdent.ReadString(const ident, default: String): String;
begin
   if (ProfileSection <> nil) and
      (ProfileSection.ProfileFile <> nil) and
      (ProfileSection.ProfileFile.GetIniFile <> nil) then
      result := ProfileSection.ProfileFile.GetIniFile.ReadString(ProfileSection.section,ident,default)
   else
      result := default;
end;

procedure TProfileIdent.WriteString(const ident, value: String);
begin
   if (ProfileSection <> nil) and
      (ProfileSection.ProfileFile <> nil) and
      (ProfileSection.ProfileFile.GetIniFile <> nil) then
      ProfileSection.ProfileFile.GetIniFile.WriteString(ProfileSection.Section,ident,value);
end;

{******************************************************************************}

constructor TProfileString.create(AOwner: TComponent);
begin
   inherited Create(Aowner);
   Update;
end;

procedure TProfileString.SetDefaultValue(v: String);
begin
   if FDefaultValue <> v then begin
      FDefaultValue := v;
      Update;
   end;
end;

procedure TProfileString.SetValue(v: String);
begin
   if FValue <> v then begin
      FValue := v;
      WriteString(Ident,v);
   end;
end;

procedure TProfileString.Update;
begin
   if not (csLoading in ComponentState) then
     FValue := ReadString(Ident,FDefaultValue);
end;

{******************************************************************************}

constructor TProfileInteger.create(AOwner: TComponent);
begin
   inherited Create(Aowner);
   Update;
end;

procedure TProfileInteger.SetDefaultValue(v: integer);
begin
   if FDefaultValue <> v then begin
      FDefaultValue := v;
      Update;
   end;
end;

procedure TProfileInteger.SetValue(v: integer);
begin
   if FValue <> v then begin
      FValue := v;
      WriteInteger(Ident,v);
   end;
end;

procedure TProfileInteger.Update;
begin
   if not (csLoading in ComponentState) then
      FValue := ReadInteger(Ident,FDefaultValue);
end;
               
{******************************************************************************}

constructor TProfileBoolean.create(AOwner: TComponent);
begin
   inherited Create(Aowner);
   Update;
end;

procedure TProfileBoolean.SetDefaultValue(v: boolean);
begin
   if FDefaultValue <> v then begin
      FDefaultValue := v;
      Update;
   end;
end;

procedure TProfileBoolean.SetValue(v: boolean);
begin
   if FValue <> v then begin
      FValue := v;
      WriteBoolean(Ident,v);
   end;
end;

procedure TProfileBoolean.Update;
begin
   if not (csLoading in ComponentState) then
      FValue := ReadBoolean(Ident,FDefaultValue);
end;

{******************************************************************************}

end.
