unit Split2;

{------------------------------------------------------------------------------}
{                                                                              }
{ TTwoWaySplitter                                                              }
{                                                                              }
{ Divides a window into two re-sizeable parts.                                 }
{ Will accept (only) two child components and divide the display between them. }
{                                                                              }
{ Freeware.  Use it at your own risk.                                          }
{                                                                              }
{ Simon Armstrong (Based on ideas from TSplitter by Graham Knight.)            }
{ 100717.3531@compuserve.com                                                   }
{                                                                              }
{ V1.1 25/2/96  Remove need for RECTS unit and minor bug fix to dashed line    }
{               around splitter bar at design time.                            }
{                                                                              }
{------------------------------------------------------------------------------}

interface

uses
  SysUtils, WinTypes, WinProcs, Classes, Controls, Graphics, Forms, ExtCtrls;

const
  DEFAULT_BAR_WIDTH = 5;
  DEFAULT_BAR_SEPARATION = 50;  { % }

type
  TSplitDirection = (sdHorizontal, sdVertical);
  TSplitPosition = procedure(Sender: TObject) of Object;
  TSplitMove = procedure(Sender: TObject; Position: Integer) of Object;
  TBarWidth = 1..MAXINT;
  TBarNumber = 0..1;
  TBarPosition = 0..100;

  TTwoWaySplitter = class(TCustomPanel)
  private
    { Private declarations }
    FSplitDirection: TSplitDirection;
    FBarPosition: TBarPosition;
    FBarColor: TColor;
    FBarWidth: TBarWidth;
    FMovingBar: TBarNumber;
    FDrawRect: TRect;
    FDrawWidth: integer;
    FDrawHeight: integer;
    FSaveCursor: TCursor;

    { events }
    FOnSplitPosition: TSplitPosition;
    FOnSplitMove: TSplitMove;

    procedure SetBarColor(const bc: TColor);
    procedure SetBarWidth(const bw: TBarWidth);
    procedure SetBarPosition(const iPos: TBarPosition);
    procedure SetSplitDirection(const sd: TSplitDirection);
    procedure DrawInverseBar;
    function  GetBarRect(const pos: TBarPosition): TRect;
    procedure PaintBar(const col: TColor; const pos: TBarPosition);
  protected
    { Protected declarations }
    constructor Create(Owner: TComponent); override;

    procedure AlignControls(AControl: TControl; var sRect: TRect); override;
    procedure CreateWnd; override;
    procedure GetDrawRect;
    procedure Loaded; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure Paint; override;
    procedure Realign;
 published
    { Published declarations }
    property Align;
    property BarColor: TColor read FBarColor write SetBarColor default clBtnFace;
    property BarPosition: TBarPosition read FBarPosition write SetBarPosition default DEFAULT_BAR_SEPARATION;
    property BarWidth: TBarWidth read FBarWidth write SetBarWidth default DEFAULT_BAR_WIDTH;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BorderStyle;
    property BorderWidth;
    property Color;
    property Ctl3D;
    property DragCursor;
    property Enabled;
    property Font;
    property Locked;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
    property SplitDirection: TSplitDirection read FSplitDirection write SetSplitDirection default sdVertical;
    property Visible;

    property OnBarPosition: TSplitPosition read FOnSplitPosition write FOnSplitPosition;
    property OnSplitMove: TSplitMove read FOnSplitMove write FOnSplitMove;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Extra', [TTwoWaySplitter]);
end;

constructor TTwoWaySplitter.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  ControlStyle := ControlStyle + [csAcceptsControls, csOpaque] - [csSetCaption];
  Width := 100;
  Height := 100;
  FMovingBar := 0;
  DragCursor := crHSplit;
  FSaveCursor := Cursor;
  FSplitDirection := sdVertical;
  FBarPosition := DEFAULT_BAR_SEPARATION;
  FBarColor := clBtnFace;
  FBarWidth := DEFAULT_BAR_WIDTH;
  Caption := '';
end;

procedure TTwoWaySplitter.CreateWnd;
begin
  inherited CreateWnd;
  GetDrawRect;
end;

procedure TTwoWaySplitter.SetBarPosition(const iPos: TBarPosition);
begin
   if (FBarPosition <> iPos) then begin
     FBarPosition := iPos;
     Realign;
   end;
end;

procedure TTwoWaySplitter.SetBarColor(const bc: TColor);
begin
   if FBarColor <> bc then begin
     FBarColor := bc;
     Refresh;
   end;
end;

procedure TTwoWaySplitter.SetSplitDirection(const sd: TSplitDirection);
begin
  if FSplitDirection <> sd then begin
    FSplitDirection := sd;
    if FSplitDirection = sdHorizontal then begin
      if DragCursor = crHSplit then
        DragCursor := crVSplit
    end else if DragCursor = crVSplit then
     DragCursor := crHSplit;
    Realign;
  end;
end;

procedure TTwoWaySplitter.SetBarWidth(const bw: TBarWidth);
begin
  if FBarWidth <> bw then begin
    FBarWidth := bw;
    Realign;
  end;
end;

procedure TTwoWaySplitter.Paint;
begin
  inherited Paint;

  with Canvas do begin
    Brush.Color := Color;
    Pen.Color := Color;
    Rectangle(FDrawRect.Left,FDrawRect.Top,FDrawRect.Right,FDrawRect.Bottom);

    PaintBar(FBarColor,FBarPosition);
  end;
end;

procedure TTwoWaySplitter.PaintBar(const col: TColor; const pos: TBarPosition);
var
  r: TRect;
  bw,i: integer;
begin
  with Canvas do begin
    if (csDesigning in ComponentState) and
       (col = Color) and
       (ControlCount < 2) then begin
       Pen.Style := psDot;     { same colour as background - draw a dotted line around it }
       Pen.Color := clBlack;
    end else begin
       Pen.Style := psSolid;
       Pen.Color := col;
    end;
    Brush.Color := col;
    r := GetBarRect(pos);

    if (BevelInner = bvLowered) or
       (BevelInner = bvRaised) then begin
       Pen.Style := psSolid;
       if BevelWidth * 2 > BarWidth then
          bw := BarWidth div 2
       else
          bw := BevelWidth;
       if BevelInner = bvLowered then
          Pen.Color := clBtnHighlight
       else
          Pen.Color := clBtnShadow;
       for i := 0 to bw -1 do begin
          if FSplitDirection = sdHorizontal then begin
             MoveTo(r.Left - i,r.Top + i);
             LineTo(r.Right + i,r.Top + i);
          end else begin
             MoveTo(r.Left + i,r.Top - i);
             LineTo(r.Left + i,r.Bottom + i);
          end;
       end;
       if BarWidth - 2 * bw > 0 then begin  { can any of the bar still be seen ? }
          Pen.Color := col;
          if FSplitDirection = sdHorizontal then
             Rectangle(r.Left - bw,r.Top + bw,r.Right + bw,r.Bottom - bw)
          else
             Rectangle(r.Left + bw,r.Top - bw,r.Right - bw,r.Bottom + bw);
       end;
       if BevelInner = bvLowered then
          Pen.Color := clBtnShadow
       else
          Pen.Color := clBtnHighlight;
       for i := 0 to bw -1 do begin
          if FSplitDirection = sdHorizontal then begin
             MoveTo(r.Left - i,r.Bottom - i - 1);
             LineTo(r.Right + i,r.Bottom - i - 1);
          end else begin
             MoveTo(r.Right - i - 1,r.Top - i);
             LineTo(r.Right - i - 1,r.Bottom + i);
          end;
       end;
    end else
       Rectangle(r.Left,r.Top,r.Right,r.Bottom);
    Pen.Style := psSolid;
  end;
end;

procedure TTwoWaySplitter.Realign;
var
  br: TRect;
begin
   if ControlCount > 0 then begin
      GetDrawRect;
      br := GetBarRect(FBarPosition);
      if FSplitDirection = sdHorizontal then
         Controls[0].SetBounds(FDrawRect.Left, FDrawRect.Top,
                               FDrawWidth, br.Top - FDrawRect.Top)
      else
         Controls[0].SetBounds(FDrawRect.Left, FDrawRect.Top,
                               br.Left - FDrawRect.Left, FDrawHeight);

      if ControlCount > 1 then begin
         if FSplitDirection = sdHorizontal then
            Controls[1].SetBounds(FDrawRect.Left, br.Bottom,
                                  FDrawWidth, FDrawRect.Bottom - br.Bottom)
         else
            Controls[1].SetBounds(br.Right, FDrawRect.Top,
                                  FDrawRect.Right - br.Right, FDrawHeight);
      end;
      Refresh;
   end;
end;

procedure TTwoWaySplitter.AlignControls(AControl: TControl; var sRect: TRect);
begin
  inherited AlignControls(AControl, sRect);
  Realign;
  if ControlCount > 1 then
     ControlStyle := ControlStyle - [csAcceptsControls]
  else
     ControlStyle := ControlStyle + [csAcceptsControls];
end;

procedure TTwoWaySplitter.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if (Button = mbLeft) and (FMovingBar = 0) then begin
    if PtInRect(GetBarRect(FBarPosition),Point(x,y)) then
      FMovingBar := 1
    else
      exit;
    if Cursor <> DragCursor then begin
       FSaveCursor := Cursor;
       Cursor := DragCursor;
    end;
    SetCapture(Handle);
    DrawInverseBar;
  end;
end;

procedure TTwoWaySplitter.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  Cursor := FSaveCursor;
  if FMovingBar <> 0 then begin
    DrawInverseBar;
    ReleaseCapture;
    Realign;
    if (Assigned(FOnSplitPosition)) then
      FOnSplitPosition(Self);
    FMovingBar := 0;
  end;
end;

procedure TTwoWaySplitter.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  iTmp, iMin, iMax, bw2, bp: Integer;
begin
  inherited MouseMove(Shift, X, Y);

  if FMovingBar <> 0 then begin
    if FSplitDirection = sdHorizontal then begin
      iTmp := Y;
      iMin := FDrawRect.Top;
      iMax := FDrawRect.Bottom;
    end else begin
      iTmp := X;
      iMin := FDrawRect.Left;
      iMax := FDrawRect.Right;
    end;

    { Check bounds of the bar }
    bw2 := FBarWidth div 2;
    if iTmp < iMin + bw2 - 1 then
      iTmp := iMin + bw2 - 1
    else if iTmp > iMax - bw2 then
      iTmp := iMax - bw2;

    bp := MulDiv(iTmp - iMin + 1, 100, iMax - iMin + 1);
    if bp = FBarPosition then
       exit;
    DrawInverseBar;
    FBarPosition := bp;
    DrawInverseBar;

    if (Assigned(FOnSplitMove)) then       { Do Move Event }
      OnSplitMove(Self, FBarPosition)
  end else if PtInRect(GetBarRect(FBarPosition),Point(x,y)) then begin
    if Cursor <> DragCursor then begin
      FSaveCursor := Cursor;
      Cursor := DragCursor;
    end;
  end else if Cursor <> FSaveCursor then
    Cursor := FSaveCursor;
end;

procedure TTwoWaySplitter.DrawInverseBar;
var
  DC: HDC;
  tl,br: TPoint;
  r: TRect;
begin
  if FMovingBar = 0 then
     exit;
  r := GetBarRect(FBarPosition);
  tl := ClientToScreen(r.TopLeft);
  br := ClientToScreen(r.BottomRight);
  DC := GetDC(0);
  BitBlt(DC, tl.X, tl.Y, br.x-tl.x, br.y-tl.y, 0, 0, 0, DSTINVERT);
  ReleaseDC(0, DC);
end;

function TTwoWaySplitter.GetBarRect(const pos: TBarPosition): TRect;
var
  p,bw2: integer;
begin
  bw2 := FBarWidth div 2;
  if FSplitDirection = sdHorizontal then begin
    p := MulDiv(pos, FDrawHeight, 100);
    result := Rect(FDrawRect.Left, FDrawRect.Top + p - bw2,
                   FDrawRect.Right, FDrawRect.Top + p - bw2 + FBarWidth);
  end else begin
    p := MulDiv(pos, FDrawWidth, 100);
    result := Rect(FDrawRect.Left + p - bw2, FDrawRect.Top,
                   FDrawRect.Left + p - bw2 + FBarWidth, FDrawRect.Bottom);
  end;
  IntersectRect(result, result, FDrawRect);
end;

procedure TTwoWaySplitter.GetDrawRect;
var
  o: integer;
begin
  o := BorderWidth;
  if BorderStyle = bsSingle then
     inc(o);
  if BevelInner <> bvNone then
     o := o + BevelWidth;
  if BevelOuter <> bvNone then
     o := o + BevelWidth;
  FDrawWidth := ClientWidth - 2*o;
  FDrawHeight := ClientHeight - 2*o;
  FDrawRect := rect(o,o,o+FDrawWidth,o+FDrawHeight);
end;

procedure TTwoWaySplitter.Loaded;
begin
  inherited Loaded;
  GetDrawRect;
end;

end.
