unit Split3;

{------------------------------------------------------------------------------}
{                                                                              }
{ TThreeWaySplitter                                                            }
{                                                                              }
{ Divides a window into three re-sizeable parts.                               }
{ Will accept (only) three child components and divide the display between     }
{ them.                                                                        }
{                                                                              }
{ Freeware.  Use it at your own risk.                                          }
{                                                                              }
{ Simon Armstrong (Based on ideas from TSplitter by Graham Knight.)            }
{ 100717.3531@compuserve.com                                                   }
{                                                                              }
{ V1.1 25/2/96  Remove need for RECTS unit and minor bug fix to dashed line    }
{               around splitter bar at design time.                            }
{------------------------------------------------------------------------------}

interface

uses
  SysUtils, WinTypes, WinProcs, Classes, Controls, Graphics, Forms, ExtCtrls;

const
  DEFAULT_BAR_WIDTH = 5;
  DEFAULT_BAR_SEPARATION = 33;  { % }

type
  TSplitDirection = (sdHorizontal, sdVertical);
  TSplitPosition = procedure(Sender: TObject) of Object;
  TSplitMove = procedure(Sender: TObject; Position: Integer) of Object;
  TBarWidth = 1..MAXINT;
  TBarNumber = 0..2;
  TBarPosition = 0..100;

  TThreeWaySplitter = class(TCustomPanel)
  private
    { Private declarations }
    FSplitDirection: TSplitDirection;
    FBarPosition1: TBarPosition;
    FBarPosition2: TBarPosition;
    FBar1Color: TColor;
    FBar2Color: TColor;
    FBarWidth: TBarWidth;
    FMovingBar: TBarNumber;
    FDrawRect: TRect;
    FDrawWidth: integer;
    FDrawHeight: integer;
    FSaveCursor: TCursor;

    { events }
    FOnSplit1Position: TSplitPosition;
    FOnSplit2Position: TSplitPosition;
    FOnSplit1Move: TSplitMove;
    FOnSplit2Move: TSplitMove;

    procedure SetBar1Color(const bc: TColor);
    procedure SetBar2Color(const bc: TColor);
    procedure SetBarWidth(const bw: TBarWidth);
    procedure SetBar1Position(const iPos: TBarPosition);
    procedure SetBar2Position(const iPos: TBarPosition);
    procedure SetSplitDirection(const sd: TSplitDirection);
    procedure DrawInverseBar;
    function  GetBarRect(const pos: TBarPosition): TRect;
    procedure PaintBar(const col: TColor; const pos: TBarPosition; const bn: TBarNumber);
  protected
    { Protected declarations }
    constructor Create(Owner: TComponent); override;

    procedure AlignControls(AControl: TControl; var sRect: TRect); override;
    procedure CreateWnd; override;
    procedure GetDrawRect;
    procedure Loaded; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure Paint; override;
    procedure Realign;
 published
    { Published declarations }
    property Align;
    property Bar1Color: TColor read FBar1Color write SetBar1Color default clBtnFace;
    property Bar2Color: TColor read FBar2Color write SetBar2Color default clBtnFace;
    property Bar1Position: TBarPosition read FBarPosition1 write SetBar1Position default DEFAULT_BAR_SEPARATION;
    property Bar2Position: TBarPosition read FBarPosition2 write SetBar2Position default 2*DEFAULT_BAR_SEPARATION;
    property BarWidth: TBarWidth read FBarWidth write SetBarWidth default DEFAULT_BAR_WIDTH;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BorderStyle;
    property BorderWidth;
    property Color;
    property Ctl3D;
    property DragCursor;
    property Enabled;
    property Font;
    property Locked;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
    property SplitDirection: TSplitDirection read FSplitDirection write SetSplitDirection default sdVertical;
    property Visible;

    property OnBar1Position: TSplitPosition read FOnSplit1Position write FOnSplit1Position;
    property OnSplit1Move: TSplitMove read FOnSplit1Move write FOnSplit1Move;
    property OnBar2Position: TSplitPosition read FOnSplit2Position write FOnSplit2Position;
    property OnSplit2Move: TSplitMove read FOnSplit2Move write FOnSplit2Move;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Extra', [TThreeWaySplitter]);
end;

constructor TThreeWaySplitter.Create(Owner: TComponent);
begin
  inherited Create(Owner);
  ControlStyle := ControlStyle + [csAcceptsControls, csOpaque] - [csSetCaption];
  Width := 100;
  Height := 100;
  FMovingBar := 0;
  DragCursor := crHSplit;
  FSaveCursor := Cursor;
  FSplitDirection := sdVertical;
  FBarPosition1 := DEFAULT_BAR_SEPARATION;
  FBarPosition2 := 2 * DEFAULT_BAR_SEPARATION;
  FBar1Color := clBtnFace;
  FBar2Color := clBtnFace;
  FBarWidth := DEFAULT_BAR_WIDTH;
  Caption := '';
end;

procedure TThreeWaySplitter.CreateWnd;
begin
  inherited CreateWnd;
  GetDrawRect;
end;

procedure TThreeWaySplitter.SetBar1Position(const iPos: TBarPosition);
begin
   if (iPos < FBarPosition2) and (FBarPosition1 <> iPos) then begin
     FBarPosition1 := iPos;
     Realign;
   end;
end;

procedure TThreeWaySplitter.SetBar2Position(const iPos: TBarPosition);
begin
   if (iPos > FBarPosition1) and (FBarPosition2 <> iPos) then begin
     FBarPosition2 := iPos;
     Realign;
   end;
end;

procedure TThreeWaySplitter.SetBar1Color(const bc: TColor);
begin
   if FBar1Color <> bc then begin
     FBar1Color := bc;
     Refresh;
   end;
end;

procedure TThreeWaySplitter.SetBar2Color(const bc: TColor);
begin
   if FBar2Color <> bc then begin
     FBar2Color := bc;
     Refresh;
   end;
end;

procedure TThreeWaySplitter.SetSplitDirection(const sd: TSplitDirection);
begin
  if FSplitDirection <> sd then begin
    FSplitDirection := sd;
    if FSplitDirection = sdHorizontal then begin
      if DragCursor = crHSplit then
        DragCursor := crVSplit
    end else if DragCursor = crVSplit then
      DragCursor := crHSplit;
    Realign;
  end;
end;

procedure TThreeWaySplitter.SetBarWidth(const bw: TBarWidth);
begin
  if FBarWidth <> bw then begin
    FBarWidth := bw;
    Realign;
  end;
end;

procedure TThreeWaySplitter.Paint;
var
  r: TRect;
begin
  inherited Paint;

  with Canvas do begin
    Pen.Color := Color;
    Brush.Color := Color;
    Rectangle(FDrawRect.Left,FDrawRect.Top,FDrawRect.Right,FDrawRect.Bottom);

    PaintBar(FBar1Color,FBarPosition1,1);
    PaintBar(FBar2Color,FBarPosition2,2);
  end;
end;

procedure TThreeWaySplitter.PaintBar(const col: TColor; const pos: TBarPosition; const bn: TBarNumber);
var
  r: TRect;
  bw,i: integer;
begin
  with Canvas do begin
    if (csDesigning in ComponentState) and
       (col = Color) and
       (ControlCount <= bn) then begin
       Pen.Style := psDot;     { same colour as background - draw a dotted line around it }
       Pen.Color := clBlack;
    end else begin
       Pen.Style := psSolid;
       Pen.Color := col;
    end;
    Brush.Color := col;
    r := GetBarRect(pos);

    if (BevelInner = bvLowered) or
       (BevelInner = bvRaised) then begin
       Pen.Style := psSolid;
       if BevelWidth * 2 > BarWidth then
          bw := BarWidth div 2
       else
          bw := BevelWidth;
       if BevelInner = bvLowered then
          Pen.Color := clBtnHighlight
       else
          Pen.Color := clBtnShadow;
       for i := 0 to bw -1 do begin
          if FSplitDirection = sdHorizontal then begin
             MoveTo(r.Left - i,r.Top + i);
             LineTo(r.Right + i,r.Top + i);
          end else begin
             MoveTo(r.Left + i,r.Top - i);
             LineTo(r.Left + i,r.Bottom + i);
          end;
       end;
       if BarWidth - 2 * bw > 0 then begin  { can any of the bar still be seen ? }
          Pen.Color := col;
          if FSplitDirection = sdHorizontal then
             Rectangle(r.Left - bw,r.Top + bw,r.Right + bw,r.Bottom - bw)
          else
             Rectangle(r.Left + bw,r.Top - bw,r.Right - bw,r.Bottom + bw);
       end;
       if BevelInner = bvLowered then
          Pen.Color := clBtnShadow
       else
          Pen.Color := clBtnHighlight;
       for i := 0 to bw -1 do begin
          if FSplitDirection = sdHorizontal then begin
             MoveTo(r.Left - i,r.Bottom - i - 1);
             LineTo(r.Right + i,r.Bottom - i - 1);
          end else begin
             MoveTo(r.Right - i - 1,r.Top - i);
             LineTo(r.Right - i - 1,r.Bottom + i);
          end;
       end;
    end else
       Rectangle(r.Left,r.Top,r.Right,r.Bottom);
    Pen.Style := psSolid;
  end;
end;

procedure TThreeWaySplitter.Realign;
var
   br1,br2: TRect;
begin
   if ControlCount > 0 then begin
      GetDrawRect;
      br1 := GetBarRect(FBarPosition1);
      if FSplitDirection = sdHorizontal then
         Controls[0].SetBounds(FDrawRect.Left, FDrawRect.Top,
                               FDrawWidth, br1.Top - FDrawRect.Top)
      else
         Controls[0].SetBounds(FDrawRect.Left, FDrawRect.Top,
                               br1.Left - FDrawRect.Left, FDrawHeight);

      if ControlCount > 1 then begin
         br2 := GetBarRect(FBarPosition2);
         if FSplitDirection = sdHorizontal then
            Controls[1].SetBounds(FDrawRect.Left, br1.Bottom,
                                  FDrawWidth, br2.Top - br1.Bottom)
         else
            Controls[1].SetBounds(br1.Right, FDrawRect.Top,
                                  br2.Left - br1.Right, FDrawHeight);

         if ControlCount > 2 then begin
            if FSplitDirection = sdHorizontal then
               Controls[2].SetBounds(FDrawRect.Left, br2.Bottom,
                                     FDrawWidth, FDrawRect.Bottom - br2.Bottom)
            else
               Controls[2].SetBounds(br2.Right, FDrawRect.Top,
                                     FDrawRect.Right - br2.Right, FDrawHeight);
         end;
      end;
      Refresh;
   end;
end;

procedure TThreeWaySplitter.AlignControls(AControl: TControl; var sRect: TRect);
begin
  inherited AlignControls(AControl, sRect);
  Realign;
  if ControlCount > 2 then
     ControlStyle := ControlStyle - [csAcceptsControls]
  else
     ControlStyle := ControlStyle + [csAcceptsControls];
end;

procedure TThreeWaySplitter.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseDown(Button, Shift, X, Y);
  if (Button = mbLeft) and (FMovingBar = 0) then begin
    if PtInRect(GetBarRect(FBarPosition1),Point(x,y)) then
      FMovingBar := 1
    else if PtInRect(GetBarRect(FBarPosition2),Point(x,y)) then
      FMovingBar := 2
    else
      exit;
    if Cursor <> DragCursor then begin
       FSaveCursor := Cursor;
       Cursor := DragCursor;
    end;
    SetCapture(Handle);
    DrawInverseBar;
  end;
end;

procedure TThreeWaySplitter.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseUp(Button, Shift, X, Y);
  Cursor := FSaveCursor;
  if FMovingBar <> 0 then begin
    DrawInverseBar;
    ReleaseCapture;
    Realign;
    if (FMovingBar = 1) and (Assigned(FOnSplit1Position)) then
      FOnSplit1Position(Self)
    else if (FMovingBar = 2) and (Assigned(FOnSplit2Position)) then
      FOnSplit2Position(Self);
    FMovingBar := 0;
  end;
end;

procedure TThreeWaySplitter.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  iTmp, iMin, iMax, bw2, bp: Integer;
begin
  inherited MouseMove(Shift, X, Y);

  if FMovingBar <> 0 then begin
    if FSplitDirection = sdHorizontal then begin
      iTmp := Y;
      iMin := FDrawRect.Top;
      iMax := FDrawRect.Bottom;
    end else begin
      iTmp := X;
      iMin := FDrawRect.Left;
      iMax := FDrawRect.Right;
    end;

    { Check bounds of the bar }
    bw2 := FBarWidth div 2;
    if iTmp < iMin + bw2 - 1 then
      iTmp := iMin + bw2 - 1
    else if iTmp > iMax - bw2 then
      iTmp := iMax - bw2;

    bp := MulDiv(iTmp - iMin + 1, 100, iMax - iMin + 1);
    if FMovingBar = 1 then begin
       if bp >= FBarPosition2 then
          bp := FBarPosition2 - 1;
       if bp = FBarPosition1 then
          exit;
       DrawInverseBar;
       FBarPosition1 := bp;
    end else begin
       if bp <= FBarPosition1 then
          bp := FBarPosition1 + 1;
       if bp = FBarPosition2 then
          exit;
       DrawInverseBar;
       FBarPosition2 := bp;
    end;
    DrawInverseBar;

    if (FMovingBar = 1) and (Assigned(FOnSplit1Move)) then       { Do Move Event }
      OnSplit1Move(Self, FBarPosition1)
    else if (FMovingBar = 2) and (Assigned(FOnSplit2Move)) then  { Do Move Event }
      OnSplit2Move(Self, FBarPosition2);
  end else if PtInRect(GetBarRect(FBarPosition1),Point(x,y)) or
              PtInRect(GetBarRect(FBarPosition2),Point(x,y)) then begin
    if Cursor <> DragCursor then begin
      FSaveCursor := Cursor;
      Cursor := DragCursor;
    end;
  end else if Cursor <> FSaveCursor then
    Cursor := FSaveCursor;
end;

procedure TThreeWaySplitter.DrawInverseBar;
var
  DC: HDC;
  tl,br: TPoint;
  r: TRect;
begin
  if FMovingBar = 1 then
     r := GetBarRect(FBarPosition1)
  else if FMovingBar = 2 then
     r := GetBarRect(FBarPosition2)
  else
     exit;
  tl := ClientToScreen(r.TopLeft);
  br := ClientToScreen(r.BottomRight);
  DC := GetDC(0);
  BitBlt(DC, tl.X, tl.Y, br.x-tl.x, br.y-tl.y, 0, 0, 0, DSTINVERT);
  ReleaseDC(0, DC);
end;

function TThreeWaySplitter.GetBarRect(const pos: TBarPosition): TRect;
var
  p,bw2: integer;
begin
  bw2 := FBarWidth div 2;
  if FSplitDirection = sdHorizontal then begin
    p := MulDiv(pos, FDrawHeight, 100);
    result := Rect(FDrawRect.Left, FDrawRect.Top + p - bw2,
                   FDrawRect.Right, FDrawRect.Top + p - bw2 + FBarWidth);
  end else begin
    p := MulDiv(pos, FDrawWidth, 100);
    result := Rect(FDrawRect.Left + p - bw2, FDrawRect.Top,
                   FDrawRect.Left + p - bw2 + FBarWidth, FDrawRect.Bottom);
  end;
  IntersectRect(result, result, FDrawRect);
end;

procedure TThreeWaySplitter.GetDrawRect;
var
  o: integer;
begin
  o := BorderWidth;
  if BorderStyle = bsSingle then
     inc(o);
  if BevelInner <> bvNone then
     o := o + BevelWidth;
  if BevelOuter <> bvNone then
     o := o + BevelWidth;
  FDrawWidth := ClientWidth - 2*o;
  FDrawHeight := ClientHeight - 2*o;
  FDrawRect := rect(o,o,o+FDrawWidth,o+FDrawHeight);
end;

procedure TThreeWaySplitter.Loaded;
begin
  inherited Loaded;
  GetDrawRect;
end;

end.
