(*

 This is an TempTable example. Free for anyone to use, modify and do
 whatever else you wish.

 This version is quite an improvement, mainly in the area of indexes.
 Both primary and secondary indexes should work just like on a regular
 TTable.

 TempTables are supposedly also in-memory tables and provide all of the
 functionality of regular tables. The problem with these is that in DB.PAS
 there is a line in TDataSet.InternalOpen that sets CanModify to False if
 the table is temporary. Why - I don't know, but if you have the VCL source
 (which I highly recommend if you're serious about Delphi programming) then
 you can just comment that part out. Go into DB.PAS, change the line in
 TDataSet.InternalOpen  that reads
    FCanModify := (CursorProps.eOpenMode = dbiReadWrite) and
    not CursorProps.bTempTable;
                               to
    FCanModify := (CursorProps.eOpenMode = dbiReadWrite); { and
    not CursorProps.bTempTable; }

 Just like all things free this unit comes with no guarantees. I cannot be
 responsible for any damage this code may cause.

 Thanks to Steve Garland <72700.2407@compuserve.com> for his help. He
 created his own variation of an in-memory table component and I used it
 to get started.

 If you have comments - please contact me at INTERNET:grisha@mira.com

 Happy hacking!

 Gregory Trubetskoy

 P.S. Here is what I had to do to overcome certain problems with Delphi's
 implementation of BDE. For a BDE novice like me, it wasn't nearly as simple
 as it would seem from reading this!

 First, when you call DbiCreateTempTable it creates a table with a different
 name. BDE uses whatever file name you give it as the beginning and appends it
 with a number, e.g. XXX.DB will become XXXn.DB where n is some number that
 gets incremented every time you call DbiCreateTempTable.

 So the first thing I did was to retrieve the new name of the table by using
 DbiGetProp and assign it to TableName. (See TTempTable.CreateTable)

 Another problem was with the indexes. When you create a primary index, you
 must switch to it with DbiSwitchToIndex to make it active. DbiSwitchToIndex
 will also return you a new cursor handle to use for the new index. Delphi
 makes sure that any value you assign to IndexValue is not the same as the old
 value, or it will do nothing. (See TTable.SetIndex in DBTABLES.PAS) Therefore,
 when you create a primary index where none existed before and IndexName is '',
 Delphi VCL does NOT call DbiSwitchToIndex to activate the new index, even if
 you do IndexName := ''. For some reason it works for regular tables (I guess
 Borland programmers know what they are doing, or I missed something), but
 for TempTables, I would get a GPF shortly after using the old cursor. So I had
 to call TDataSet.SwitchToIndex manually (See TTempTable.AddIndex).

 Last problem I had was with either Delphi VLC or BDE itself confused with
 what the file name for the table was. I would get 'Table does not exist' or
 'Operation not applicable' errors when adding indexes. I didn't study the
 problem in detail, but found it can easily be solved by DbiMakePermanent
 function (See TTempTable.Create). Description of MakePermanent sais that it
 doesn't save the table until convenient or the cursor is closed, so no
 benefit of TTempTable being fast because it's in RAM should be lost. I simply
 delete the table in the destructor to reverse the effect of DbiMakePermanent.

 Set the Permanent property to True, to prevent the table from being deleted.

 Speaking of performance benefits - my test showed that writing and indexing
 5000 small records takes 5.77 seconds with TTable and 5.40 seconds with
 TTempTable - so I'm not sure where the performance benefit is.

 I do think TempTable is highly useful, mainly because it's a neat way to
 manage temporary tables without having to do all the boring stuff, like
 creating temporary filenames, if not for the supposed performance improvement.

 Happy Hacking again, and make sure to drop me a thank you note if you're using
 this!

 My e-mail is <grisha@mira.com> (without brackets, of course)

*)

unit Temptbl;

interface

uses Classes, DBTables, WinTypes, WinProcs, DBITypes, DBIProcs, DB, SysUtils;


type TTempTable = class(TTable)
  private
    hCursor: hDBICur;
    FPermanent: Boolean;
    procedure EncodeFieldDesc(var FieldDesc: FLDDesc;
      const Name: string; DataType: TFieldType; Size: Word);
    function CreateHandle: HDBICur; override;
    procedure EncodeIndexDesc(var IndexDesc: IDXDesc;
      const Name, Fields: string; Options: TIndexOptions);
  public
    destructor Destroy; override;
    constructor Create(AOwner: TComponent);
    procedure CreateTable;
    procedure AddIndex(const Name, Fields: string; Options: TIndexOptions);
    property Permanent: Boolean read FPermanent write FPermanent;
  end;

implementation

constructor TTempTable.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Permanent := False;
end;

destructor TTempTable.Destroy;
begin
  if not Permanent then begin
    Close;
    DeleteTable;
  end;
  inherited Destroy;
end;

procedure TTempTable.AddIndex(const Name, Fields: string; Options: TIndexOptions);
var
  SIndexName, STagName: array[0..255] of char;
begin
  inherited AddIndex(Name, Fields, Options);
  if ixPrimary in Options then begin
    StrPCopy(SIndexName, ''); { always blank for ixPrimary }
    StrPCopy(STagName, '');
    SwitchToIndex(SIndexName, STagName);
  end;
end;

function TTempTable.CreateHandle;
begin
  Result := hCursor;
end;

procedure TTempTable.EncodeIndexDesc(var IndexDesc: IDXDesc;
  const Name, Fields: string; Options: TIndexOptions);
var
  Pos: Integer;
begin
  FillChar(IndexDesc, SizeOf(IndexDesc), 0);
  with IndexDesc do
  begin
  {  if IsDBaseTable then
      AnsiToNative(DBLocale, Name, szTagName, SizeOf(szTagName) - 1)
    else
      AnsiToNative(DBLocale, Name, szName, SizeOf(szName) - 1);   }
    bPrimary := ixPrimary in Options;
    bUnique := ixUnique in Options;
    bDescending := ixDescending in Options;
    bMaintained := True;
    bCaseInsensitive := ixCaseInsensitive in Options;
    if ixExpression in Options then
    begin
      bExpIdx := True;
      AnsiToNative(DBLocale, Fields, szKeyExp, SizeOf(szKeyExp) - 1);
    end else
    begin
      Pos := 1;
      while (Pos <= Length(Fields)) and (iFldsInKey < 16) do
      begin
        aiKeyFld[iFldsInKey] :=
          FieldDefs.Find(ExtractFieldName(Fields, Pos)).FieldNo;
        Inc(iFldsInKey);
      end;
    end;
  end;
end;


procedure TTempTable.EncodeFieldDesc(var FieldDesc: FLDDesc;
  const Name: string; DataType: TFieldType; Size: Word);
const
  TypeMap: array[TFieldType] of Byte = (
    fldUNKNOWN, fldZSTRING, fldINT16, fldINT32, fldUINT16, fldBOOL,
    fldFLOAT, fldFLOAT, fldBCD, fldDATE, fldTIME, fldTIMESTAMP, fldBYTES,
    fldVARBYTES, fldBLOB, fldBLOB, fldBLOB);
begin
  with FieldDesc do
  begin
    AnsiToNative(Locale, Name, szName, SizeOf(szName) - 1);
    iFldType := TypeMap[DataType];
    case DataType of
      ftString, ftBytes, ftVarBytes, ftBlob, ftMemo, ftGraphic:
        iUnits1 := Size;
      ftBCD:
        begin
          iUnits1 := 32;
          iUnits2 := Size;
        end;
    end;
    case DataType of
      ftCurrency:
        iSubType := fldstMONEY;
      ftBlob:
        iSubType := fldstBINARY;
      ftMemo:
        iSubType := fldstMEMO;
      ftGraphic:
        iSubType := fldstGRAPHIC;
    end;
  end;
end;

procedure TTempTable.CreateTable;
var
  I, J: Integer;
  FieldDescs: PFLDDesc;
  ValCheckPtr: PVCHKDesc;
  DriverTypeName: DBINAME;
  TableDesc: CRTblDesc;
  TempTableName: pChar;
  TempTableNameLen: Word;
begin
  CheckInactive;
  if FieldDefs.Count = 0 then
    for I := 0 to FieldCount - 1 do
      with Fields[I] do
        if not Calculated then
          FieldDefs.Add(FieldName, DataType, Size, Required);
  FieldDescs := nil;
  FillChar(TableDesc, SizeOf(TableDesc), 0);
  with TableDesc do
  begin
    SetDBFlag(dbfTable, True);
    try
      AnsiToNative(Locale, TableName, szTblName, SizeOf(szTblName) - 1);
      {if GetTableTypeName <> nil then
        StrCopy(szTblType, GetTableTypeName);}
      iFldCount := FieldDefs.Count;
      FieldDescs := AllocMem(iFldCount * SizeOf(FLDDesc));
      for I := 0 to FieldDefs.Count - 1 do
        with FieldDefs[I] do
        begin
          EncodeFieldDesc(PFieldDescList(FieldDescs)^[I], Name,
            DataType, Size);
          if Required then Inc(iValChkCount);
        end;
      pFldDesc := AllocMem(iFldCount * SizeOf(FLDDesc));
      Check(DbiTranslateRecordStructure(nil, iFldCount, FieldDescs,
        nil, nil, pFLDDesc));
      iIdxCount := IndexDefs.Count;
      pIdxDesc := AllocMem(iIdxCount * SizeOf(IDXDesc));
      for I := 0 to IndexDefs.Count - 1 do
        with IndexDefs[I] do
          EncodeIndexDesc(PIndexDescList(pIdxDesc)^[I], Name, Fields,
            Options);
      if iValChkCount <> 0 then
      begin
        pVChkDesc := AllocMem(iValChkCount * SizeOf(VCHKDesc));
        ValCheckPtr := pVChkDesc;
        for I := 0 to FieldDefs.Count - 1 do
          if FieldDefs[I].Required then
          begin
            ValCheckPtr^.iFldNum := I + 1;
            ValCheckPtr^.bRequired := True;
            Inc(ValCheckPtr);
          end;
      end;
      Check(DbiCreateTempTable(DBHandle, TableDesc, hCursor));
      Check(DbiSetProp(hDBIObj(hCursor), curXLTMODE, LongInt(xltFIELD)));
      GetMem(TempTableName, 256);
      Check(DbiGetProp(hDBIObj(hCursor), curFILENAME, TempTableName, 256, TempTableNameLen));
      TableName := StrPas(TempTableName);
      Check(DBIMakePermanent(hCursor, nil, True));
    finally
      if pVChkDesc <> nil then FreeMem(pVChkDesc, iValChkCount * SizeOf(VCHKDesc));
      if pIdxDesc <> nil then FreeMem(pIdxDesc, iIdxCount * SizeOf(IDXDesc));
      if pFldDesc <> nil then FreeMem(pFldDesc, iFldCount * SizeOf(FLDDesc));
      if FieldDescs <> nil then FreeMem(FieldDescs, iFldCount * SizeOf(FLDDesc));
      SetDBFlag(dbfTable, False);
      FreeMem(TempTableName, 256);
    end;
  end;
end;

end.
