unit ColorEd;

{ This unit declares the TRgbColorProperty editor. This property editor
  is similar to TColorProperty, the default color property editor, but
  it adds three subproperties for editting the constituent RGB values
  without opening the Windows color chooser dialog box.

  It is up to the reader to decide whether to use this unit and its property
  editor. The primary purpose for this unit is instructional: to demonstrate
  how to add subproperties to any property type.

  For reasons unknown to the author, Borland decided to make it impossible
  for the programmer to create subproperties for an arbitrary property editor.
  This privilege is reserved to TClassProperty and TSetProperty. No property
  editor that is declared outside of DsgnIntf.pas can create an instance of
  any property editor class. The only public routine that creates property
  editors is GetComponentProperties. Hence, the only way to create a property
  editor object is to have one or more components and create their properties.

  Therein lies the trick to TRgbColorProperty. It creates special, hidden
  objects that have three published properties: Red, Green, and Blue. When the
  user double clicks the Color property, to expose the subproperties, the
  Object Inspector calls the TRgbColorProperty.GetProperties method. This
  method, in turn, creates the hidden, TRgbKludge objects, and calls
  GetComponentProperties to allocate the Red, Green, and Blue property editors.

  Note that Delphi automatically sorts all property editors into alphabetical
  order. This can be circumvented, but is the subject of another lesson. Hence,
  the user sees the properties listed in the order: Blue, Green, Red.

  Ray Lischner
  Tempest Software
  2 January 1996

  Copyright  1996 Tempest Software.
  Permission is granted to use this software in any application or
  similar program, without fee or royalty. Please share this software
  with your friends and colleagues, and feel free to improve it
  and modify it to suit your needs, but leave this copyright notice intact.
}

interface

uses
  Classes, SysUtils, Graphics, DsgnIntf;

  type
    { The master property editor is TRgbColorProperty. It inherits from
      TColorProperty, to inherit the drop-down list of color names,
      and the edit dialog box. It adds only subproperties. }
    TRgbColorProperty = class(TColorProperty)
    private
      ChildList: TComponentList;  { list of hidden child components }
      procedure SetChildColors;
    public
      destructor Destroy; override;
      function GetAttributes: TPropertyAttributes; override;
      procedure GetProperties(Proc: TGetPropEditProc); override;
      procedure SetValue(const Value: string); override;
      procedure SetColor(Value: TColor);
    end;

    { TRgbColorProperty creates a TRgbKludge componentobject as a companion
      to each selected component. It defines the Red, Green, and Blue
      properties which become the subproperties of TRgbColorProperty.
      When any of the constituent color parts changes, TRgbKludge
      notifies the TRgbColorProperty editor, so it can update
      the TColor value. Remember that when a property value is set,
      the same value is set for all selected components.

      The RGB values are stored as Byte values. Thus, white is
      (255, 255, 255), and Black is (0,0,0). This allows the easiest
      way to create a TColor value, by using the RGB function.

      It is conceivable, however, that one might prefer, say, a 1..100
      scale, or evey a floating point value between 0 and 1. I have
      also seen a 16-bit value used, although you lose that extra
      precision when converting to the TColor value. }
    TRgbValue = Byte;
    TRgbKludge = class(TPersistent)
    private
      fColor: TColor;   { Cache the current color value, for convenience. }
      fEditor: TRgbColorProperty; { pointer back to the master property editor }
      function GetRed: TRgbValue;
      procedure SetRed(Value: TRgbValue);
      function GetGreen: TRgbValue;
      procedure SetGreen(Value: TRgbValue);
      function GetBlue: TRgbValue;
      procedure SetBlue(Value: TRgbValue);
      procedure SetColor(Value: TColor);
    public
      property Color: TColor read fColor write SetColor;
      property Editor: TRgbColorProperty read fEditor write fEditor;
    published
      property Red: TRgbValue read GetRed write SetRed;
      property Green: TRgbValue read GetGreen write SetGreen;
      property Blue: TRgbValue read GetBlue write SetBlue;
    end;

  procedure Register;

implementation

uses WinProcs, TypInfo;

destructor TRgbColorProperty.Destroy;
begin
  ChildList.Free;
  inherited Destroy;
end;

function TRgbColorProperty.GetAttributes: TPropertyAttributes;
begin
  Result := inherited GetAttributes + [paSubProperties];
end;

{ Update all the hidden, child property editors, if any. }
procedure TRgbColorProperty.SetChildColors;
var
  I: Integer;
begin
  if ChildList <> nil then
    for I := 0 to ChildList.Count-1 do
        TRgbKludge(ChildList[I]).Color := GetOrdValueAt(I);
end;

{ When the parent's color changes, make sure all the child
  objects change to match. }
procedure TRgbColorProperty.SetValue(const Value: string);
begin
  inherited SetValue(Value);
  SetChildColors;
end;

procedure TRgbColorProperty.SetColor(Value: TColor);
begin
  SetOrdValue(Value);
  SetChildColors;
end;

{ When the Object Inspector requests the subproperties, it is time
  for TRgbColorProperty to do its thing. Create a hidden TRgbKludge
  object to parallel each component that is currently selected.
  Note that TComponentList is intended to hold a list of components,
  but TRgbKludge is just a TPersistent subclass. This causes no
  harm, but it requires a few explicit type casts. Future versions
  of Delphi might cause this to be aproblem, in which case you would
  need to derive TRgbKludge from TComponent, and filter out the
  Name and Tag propertis. }
procedure TRgbColorProperty.GetProperties(Proc: TGetPropEditProc);
var
  RgbKludge: TRgbKludge;
  I: Integer;
begin
  if ChildList <> nil then
    SetChildColors
  else
  begin
    { Create the list of hidden objects. }
    ChildList := TComponentList.Create;
    for I := 0 to PropCount-1 do
    begin
      RgbKludge := TRgbKludge.Create;
      RgbKludge.Color := GetOrdValueAt(I);
      Rgbkludge.Editor := Self;
      ChildList.Add(TComponent(RgbKludge));
    end;
  end;

  { Get the subproperty editors. }
  GetComponentProperties(ChildList, tkProperties, Designer, Proc);
end;


{ TRgbKludge }

{ Set the color and notify the master property editor. }
procedure TRgbKludge.SetColor(Value: TColor);
begin
  if fColor <> Value then
  begin
    fColor := Value;
    if Editor <> nil then
      Editor.SetColor(Value);
  end;
end;

function TRgbKludge.GetRed: TRgbValue;
begin
  Result := GetRValue(ColorToRgb(Color))
end;

procedure TRgbKludge.SetRed(Value: TRgbValue);
begin
  Color := RGB(Value, Green, Blue)
end;

function TRgbKludge.GetGreen: TRgbValue;
begin
  Result := GetGValue(ColorToRgb(Color))
end;

procedure TRgbKludge.SetGreen(Value: TRgbValue);
begin
  Color := RGB(Red, Value, Blue)
end;

function TRgbKludge.GetBlue: TRgbValue;
begin
  Result := GetBValue(ColorToRgb(Color))
end;

procedure TRgbKludge.SetBlue(Value: TRgbValue);
begin
  Color := RGB(Red, Green, Value)
end;


procedure Register;
begin
  RegisterPropertyEditor(TypeInfo(TColor), nil, '', TRgbColorProperty);
end;

end.
